/*
** Module   : ProcedureManager.java
** Abstract : Implements access to procedure system handles and procedure
**            related attributes and methods. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CA  20130112 Created initial version.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.security.*;

/**
 * Provides implementation for the following procedure handles:
 * <table>
 * <tr>
 *   <td>THIS-PROCEDURE</td>
 *   <td>{@link #thisProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>SOURCE-PROCEDURE</td>
 *   <td>{@link #sourceProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>TARGET-PROCEDURE</td>
 *   <td>{@link #targetProcedure()}</td>
 * </tr>
 * </table>
 * The following procedure-related attributes and methods are implemented:
 * <table>
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>SESSION:SUPER-PROCEDURES</td>
 *    <td>{@link #superProcedures()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:ADD-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #addSuperProcedure(handle)}</li>
 *          <li>{@link #addSuperProcedure(handle, SearchMode)}</li>
 *          <li>{@link #addSuperProcedure(handle, String)}</li>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:REMOVE-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #removeSuperProcedure(handle)}</li>
 *       </ul>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:FIRST-PROCEDURE</td>
 *    <td>{@link #firstProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:LAST-PROCEDURE</td>
 *    <td>{@link #lastProcedure()}</td>
 * </tr>
 * </table>
 * <p>
 * The {@link #registerFunctionHandle(String, Resolvable)} is emitted at the
 * beginning of an external procedure, for any function defined 
 * <code>IN handle</code>.
 */
public class ProcedureManager
{
   /** 
    * Defines search modes for the ADD-SUPER-PROCEDURE statement. Needed to
    * be able to identify the correct target in RUN SUPER and SUPER() cases.
    */
   public static enum SearchMode
   {
      SEARCH_SELF
      {
         @Override
         public String toString()
         {
            return "SEARCH-SELF";
         }
      },
      
      SEARCH_TARGET
      {
         @Override
         public String toString()
         {
            return "SEARCH-TARGET-PROCEDURE";
         }
      }
   };

   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();

   /**
    * Register the handle expression for the given function.
    * 
    * @param    pname
    *           The function's legacy 4GL name.
    * @param    handleExpr
    *           The {@link Resolvable} instance used to determine its IN 
    *           handle target.
    */
   public static void registerFunctionHandle(String pname, 
                                             Resolvable handleExpr)
   {
      // always work with thisProcedure
      handle h = thisProcedure();
      Object referent = h.get();
      PersistentProc pp = work.get().ppMap.get(referent);
      pp.inHandleFunctions.put(pname, handleExpr);
   }
   
   /**
    * Get the THIS-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle thisProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.thisProcedures.peek();
      return new handle(wa.ppMap.get(referent).phandle);
   }
   
   /**
    * Get the TARGET-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle targetProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.peek();
      return new handle(wa.ppMap.get(referent).phandle);
   }

   /**
    * Get the SOURCE-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle sourceProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.sourceProcedures.peek();

      return referent == null ? new handle()
                              : new handle(wa.ppMap.get(referent).phandle);
   }
   
   /**
    * Get the SESSION:LAST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle lastProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }

      Object referent = wa.persistentProcedures.peekLast();
      return new handle(wa.ppMap.get(referent).phandle);
   }
   
   /**
    * Get the SESSION:FIRST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle firstProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }

      Object referent = wa.persistentProcedures.peekFirst();     
      return new handle(wa.ppMap.get(referent).phandle);
   }
   
   /**
    * Get the SESSION:SUPER-PROCEDURES attribute.
    * 
    * @return   See above.
    */
   public static character superProcedures()
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSessionSuperProcedures().values());
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h)
   {
      return addSuperProcedure(h, (SearchMode) null);
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, String searchMode)
   {
      return addSuperProcedure(h, getSearchMode(searchMode));
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, null);
   }

   /**
    * Remove the given handle from the SESSION's list of super-procedures.
    *  
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical removeSuperProcedure(handle h)
   {
      if (!h._isValid() || !h.isProcedure())
      {
         return new logical(false);
      }

      removeProcedureImpl(h, work.get().getSessionSuperProcedures().keySet());
   
      return new logical(true);
   }

   /**
    * Ensure all the context-local data has initialized (to be able to listen
    * for scope notifications).
    */
   static void load()
   {
      // needed to ensure the variable is loaded, so it listens for scope
      // notifications
      work.get();
   }
   
   /**
    * Add the <code>h</code> handle as a persistent procedure. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)}.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    */
   static void addPersistentProcedure(handle h)
   {
      WorkArea wa = work.get();
      
      PersistentProc pp = 
         removeProcedureImpl(h, wa.persistentProcedures);
      if (pp == null)
      {
         pp = new PersistentProc(h, null);
         wa.ppMap.put(pp.phandle.get(), pp);
      }

      wa.persistentProcedures.add(pp.phandle.get());
   }

   /**
    * Remove the given handle from the list of persistent procedures. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)} in
    * cases when errors were encountered while invoking the procedure.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    */
   static void removePersistentProcedure(handle h)
   {
      removeProcedureImpl(h, work.get().persistentProcedures);
   }

   /**
    * Get the FILE-NAME attribute of the given procedure handle, passed as
    * the external procedure instance. 
    * Used by {@link ExternalProgramWrapper#getFileName()} and other internal
    * APIs.
    * 
    * @param    referent
    *           The referent for which the name is needed.
    *           
    * @return   The FILE-NAME attribute.
    */
   static character getFileName(Object referent)
   {
      String jname = referent.getClass().getName();
      String pname = SourceNameMapper.convertJavaProg(jname);
      return new character(pname);
   }

   /**
    * Check if the given referent has been (or is still) registered with the
    * {@link ProcedureManager} in the {@link WorkArea#ppMap persistent 
    * procedure} map.
    * 
    * @param     referent
    *            The referent to be checked if is still loaded.
    * 
    * @return    <code>true</code> if the referent was found.
    */
   static boolean hasReferent(Object referent)
   {
      return work.get().ppMap.containsKey(referent);
   }
   
   /**
    * Delete the given procedure handle (i.e. execute the DELETE OBJECT or
    * DELETE PROCEDURE handle). At this call, the handle is known to be a
    * valid procedure handle.
    * 
    * @param    h
    *           A valid procedure handle.
    */
   static void delete(handle h)
   {
      // handle is always valid
      WorkArea wa = work.get();
      Object referent = h.get();
      
      wa.persistentProcedures.remove(referent);
      wa.sessionSuperProcs.remove(referent);
      
      for (PersistentProc pp : wa.ppMap.values())
      {
         pp.superProcedures.remove(referent);
      }

      wa.cleanup(referent);
   }
   
   /**
    * Push the new SOURCE-PROCEDURE onto the {@link WorkArea#sourceProcedures}
    * stack (this pushes the current THIS-PROCEDURE, as this is invoked before
    * a function or procedure call).
    */
   static void pushSourceProcedure()
   {
      handle h = thisProcedure();
      work.get().sourceProcedures.push(h.get());
   }

   /**
    * Pop the SOURCE-PROCEDURE from {@link WorkArea#sourceProcedures} stack.
    */
   static void popSourceProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.sourceProcedures.pop();
      wa.cleanup(referent);
   }

   /**
    * Push the given handle as the new TARGET-PROCEDURE on the 
    * {@link WorkArea#targetProcedures} stack.
    * 
    * @param    h
    *           The new TARGET-PROCEDURE.
    */
   static void pushTargetProcedure(handle h)
   {
      WorkArea wa = work.get();
      Object referent = h.get();
      wa.targetProcedures.push(referent);
   }

   /**
    * Pop the TARGET-PROCEDURE from the stack.
    */
   static void popTargetProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.pop();
      wa.cleanup(referent);
   }

   /**
    * Push the given handle as a used super-procedure, for this-procedure.
    * 
    * @param    h
    *           The used super-procedure.
    */
   static void pushSuperProcedure(handle h)
   {
      WorkArea wa = work.get();
      Object referent = thisProcedure().get();
      wa.ppMap.get(referent).usedSuperProcedures.push(h.get());
   }

   /**
    * Push the last used super-procedure, for this-procedure.
    */
   static void popSuperProcedure()
   {
      WorkArea wa = work.get();
      Object referent = thisProcedure().get();
      wa.ppMap.get(referent).usedSuperProcedures.pop();
   }

   /**
    * Using the name of a function defined IN handle and a procedure handle 
    * to which this function definition belongs, resolve the target handle,
    * to be able to invoke it.
    * 
    * @param    phandle
    *           The procedure handle where the function was defined.
    * @param    fname
    *           The function's legacy name.
    *           
    * @return   The procedure handle where to search for the function's body.
    */
   static handle getHandleForFunction(handle phandle, String fname)
   {
      Object referent = phandle.get();
      PersistentProc pp = work.get().ppMap.get(referent);
   
      handle h = (handle) pp.inHandleFunctions.get(fname).resolve();
      return h == null ? new handle() : h;
   }

   /**
    * For procedures or functions defined IN SUPER, this computes the target
    * procedure handle where the procedure's or function's body resides. 
    * 
    * @param    iename
    *           Internal-entry's name (this may be a procedure or function).
    * @param    function
    *           <code>true</code> if this is a function.
    *           
    * @return   A procedure handle where to search for the internal-entry's
    *           body.
    */
   static handle getSuperHandle(String iename, boolean function)
   {
      WorkArea wa = work.get();
      
      handle hthis   = thisProcedure();
      handle htarget = targetProcedure();
      PersistentProc ppThis   = wa.ppMap.get(hthis.get());
      PersistentProc ppTarget = wa.ppMap.get(htarget.get());
   
      Collection<PersistentProc>[] search = new Collection[2];
      search[0] = null;
      search[1] = wa.sessionSuperProcs.values();
      
      // walk the list of this-procedure's
      boolean searchSelf = ppThis.isSearchSelf();
      if (searchSelf)
      {
         search[0] = ppThis.superProcedures.values();
      }
      else
      {
         search[0] = ppTarget.superProcedures.values();
      }
      
      for (int i = 0; i < search.length; i++)
      {
         // search is done backwards (last set super-procedure is searched
         // first)
         ArrayDeque<PersistentProc> queue = 
            new ArrayDeque<PersistentProc>(search[i]);
         while (!queue.isEmpty())
         {
            PersistentProc pp = queue.removeLast();
            // skip over already used super-procedures, when in search-target
            // mode
            if (!searchSelf && 
                ppTarget.usedSuperProcedures.contains(pp.phandle.get()))
            {
               continue;
            }

            character pname = pp.pname;
            String spname = pname.toStringMessage();
            String jname = SourceNameMapper.getMethodName(spname, 
                                                          iename,
                                                          function);
            if (jname != null)
            {
               // found it, nothing else to do
               return pp.phandle; 
            }
         }
      }
      
      return new handle();
   }

   /**
    * Check if the given referent is for a persistent procedure. This 
    * implements the PERSISTENT attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   static logical isPersistent(Object referent)
   {
      return new logical(work.get().persistentProcedures.contains(referent));
   }

   /**
    * Get the <code>SUPER-PROCEDURES</code> attribute of the specified 
    * external procedure instance.
    *
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character superProcedures(Object referent)
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSuperProcedures(referent).values());
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h)
   {
      return addSuperProcedure(referent, h, (SearchMode) null);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the explicit {@link SearchMode mode}.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h,SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, referent);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the search-mode set as a string (one of the 
    * <code>SEARCH-TARGET</code> or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h, String mode)
   {
      return addSuperProcedure(referent, h, getSearchMode(mode));
   }

   /**
    * Remove the given handle from the specified referent's list of 
    * super-procedures.
    *  
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical removeSuperProcedure(Object referent, handle h)
   {
      if (!h._isValid() || !h.isProcedure())
      {
         return new logical(false);
      }

      removeProcedureImpl(h, work.get()
                                 .getSuperProcedures(referent)
                                 .keySet());

      return new logical(true);
   }
   
   /**
    * Get the <code>NEXT-SIBLING</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getNextSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = null;
      Object next = null;
      for (Object pp : wa.persistentProcedures)
      {
         if (prev != null && prev == referent)
         {
            next = pp;
            break;
         }

         prev = pp;
      }
      
      if (next == null)
      {
         return new handle();
      }

      return wa.ppMap.get(next).phandle;
   }
   
   /**
    * Get the <code>PREV-SIBLING</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getPrevSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = null;
      for (Object pp : wa.persistentProcedures)
      {
         if (pp == referent)
         {
            break;
         }
         prev = pp;
      }
      
      if (prev == null)
      {
         return new handle();
      }

      return wa.ppMap.get(prev).phandle;
   }

   static void setPrivateData(Object referent, character data)
   {
      PersistentProc pp = work.get().ppMap.get(referent);
      pp.privateData = new character(data);
   }

   /**
    * Get the <code>PRIVATE-DATA</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static character getPrivateData(Object referent)
   {
      PersistentProc pp = work.get().ppMap.get(referent);

      return pp.privateData == null ? new character() 
                                    : new character(pp.privateData);
   }

   /**
    * Get a handle list of super-procedures for the given referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures(Object referent)
   {
      WorkArea wa = work.get();
      
      List<handle> superProcs = new LinkedList<handle>();
      // need to emit them in reverse order (they are kept in LIFO)
      for (PersistentProc pp : wa.getSuperProcedures(referent).values())
      {
         superProcs.add(0, pp.phandle);
      }
      
      return superProcs;
   }
   
   /**
    * Get a handle list of super-procedures for the session handle.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures()
   {
      return getSuperProcedures(null);
   }
   
   /**
    * Check if the given referent is for a system handle (i.e. for one of the
    * THIS-PROCEDURE, SOURCE-PROCEDURE or TARGET-PROCEDURE handles).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static boolean systemHandle(Object referent)
   {
      WorkArea wa = work.get();
      return wa.thisProcedures.contains(referent)   ||
             wa.targetProcedures.contains(referent) ||
             wa.sourceProcedures.contains(referent);
   }

   /**
    * Compute the THIS-PROCEDURE handle using {@link Block} instance and the 
    * <code>this$</code> fields added by the compiler for each instance inner
    * class. When <code>this$0</code> is reached, it returns the handle for
    * that instance.
    *   
    * @return   See above.
    */
   private static handle thisProcedureImpl()
   {
      // this is the current executing procedure. need to determine the first
      // Block instance in the stack and after that resolve this$0 instance
      BlockDefinition block = TransactionManager.nearestExecutingBlock();

      // in hand written-code, we can end up with executing blocks being used
      // in static methods, called outside a real external procedure instance
      // in such cases, the Block instance has no this$ field; in such cases,
      // it is assumed that the previous THIS-PROCEDURE is the correct one,
      // so we return the top of the thisProcedures stack
      Field[] fields = block.block.getClass().getDeclaredFields();
      boolean foundThis = false;
      for (int i = 0; i < fields.length; i++)
      {
         Field f = fields[i];
         String fname = f.getName();
         if (fname.startsWith("this$"))
         {
            foundThis = true;
            break;
         }
      }
      Object outer = null;
      if (!foundThis)
      {
         outer = work.get().thisProcedures.peek();
      }
      else
      {
         outer = findRootEnclosingInstance(block.block);
         if (outer == null)
         {
            throw new RuntimeException("Could not determine THIS-PROCEDURE");
         }
      }
      
      handle phandle = new handle();
      
      ExternalProgramWrapper extProg = new ExternalProgramWrapper(outer);
      phandle.set(extProg);
      
      return phandle;
   }

   /**
    * Find the root enclosing instance of the specified inner class instance.
    * It gets computed using the <code>this$</code> fields inserted by the 
    * compiler; it returns the instance refered by the <code>this$0</code> 
    * fields, when reached.
    * 
    * @param    instance
    *           The inner class instance.
    *           
    * @return   The root instance.
    */
   private static Object findRootEnclosingInstance(Object instance)
   {
      Field[] fields = instance.getClass().getDeclaredFields();
      for (int i = 0; i < fields.length; i++)
      {
         Field f = fields[i];
         String fname = f.getName();
         if (fname.startsWith("this$"))
         {
            try
            {
               f.setAccessible(true);
               Object outer = f.get(instance);
               
               return 
                  fname.equals("this$0") ? outer 
                                         : findRootEnclosingInstance(outer);
            }
            catch (Exception e)
            {
               e.printStackTrace();
               return null;
            }
         }
      }
      
      return null;
   }

   /**
    * Implementation for the SUPER-PROCEDURES attribute.
    *  
    * @param    superProcs
    *           A list of {@link PersistentProc} objects, in the order in 
    *           which they were added.
    *           
    * @return   The string representation of each super-procedure, separated
    *           by a comma. The last added super-procedure must be first in
    *           this list, the second-to-last must be the second in the list,
    *           etc.
    */
   private static character superProceduresImpl(
                                        Collection<PersistentProc> superProcs)
   {
      character sp = new character("");
      boolean first = true;
      ArrayDeque<PersistentProc> queue = new ArrayDeque<PersistentProc>();
      queue.addAll(superProcs);
      // need to emit them in a LIFO order
      while (!queue.isEmpty())
      {
         PersistentProc pp = queue.removeLast();
         int hashCode = pp.phandle.hashCode();
         sp.assign(character.concat(sp, (!first ? "," : ""), hashCode));
         first = false;
      }
      
      return sp;
   }
   
   /**
    * The implementation of the ADD-SUPER-PROCEDURE method. When the
    * <code>referent</code> is null, this call is for the SESSION handle.
    * 
    * @param    h
    *           A handle to be added as super-procedure.
    * @param    mode
    *           The {@link SearchMode} for the passed handle.
    * @param    referent
    *           When <code>null</code> add to the SESSION:SUPER-PROCEDURES
    *           list. Else, add to the procedure associated with this 
    *           referent.
    * 
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   private static logical addSuperProcedureImpl(handle h, 
                                                SearchMode mode,
                                                Object referent)
   {
      if (!h._isValid() || !h.isProcedure())
      {
         return new logical(false);
      }
      
      WorkArea wa = work.get();
      Map<Object, PersistentProc> superProcs = 
         wa.getSuperProcedures(referent);
      PersistentProc oldPp = removeProcedureImpl(h, superProcs.keySet());
      PersistentProc newPp = null;

      if (mode == null             ||
          oldPp == null            || 
          oldPp.searchMode == null ||  
          oldPp.searchMode.equals(mode))
      {
         newPp = oldPp == null ? new PersistentProc(h, mode) : oldPp;
         newPp.searchMode = mode;
         wa.ppMap.put(newPp.phandle.get(), newPp);
         superProcs = wa.getSuperProcedures(referent);         
      }
      else
      {
         newPp = oldPp;

         int errNum = 8914;
         String msg = 
            "**Changing proc-search for procedure %s from  '%s' to '%s'";

         String progName = oldPp.pname.toStringMessage();
         String oldMode  = oldPp.searchMode.toString();
         String newMode  = mode.toString();
         
         msg = String.format(msg, progName, oldMode, newMode);

         // this is just a warning, the mode will be changed
         ErrorManager.displayError(errNum, msg, false);
      }
      
      if (referent != null)
      {
         // TODO: if we've started a SUPER call in SEARCH-TARGET mode and
         // the h handle is one of not-yet-used super-procedures for the
         // TARGET-PROCEDURE handle, than all future SUPER calls will fail.
         // need to invalidate them.
      }

      superProcs.put(newPp.phandle.get(), newPp);

      return new logical(true);
   }

   /**
    * Removes the given procedure handle from the given procedure instances
    * list.
    * 
    * @param    h
    *           The procedure handle to be removed.
    * @param    ppList
    *           The collection from which the procedure must be removed.
    *           
    * @return   The {@link PersistentProc} instance, if it was found in the
    *           <code>ppList</code> list.
    */
   private static PersistentProc removeProcedureImpl(handle h, 
                                                    Collection<Object> ppList)
   {
      WorkArea wa = work.get();
      
      Object hreferent = h.get();
      PersistentProc pp = null;
      
      Iterator<Object> iter = ppList.iterator();
      while (iter.hasNext())
      {
         Object referent = iter.next();
         if (hreferent == referent)
         {
            iter.remove();
            pp = wa.ppMap.get(hreferent);
         }
      }

      wa.cleanup(hreferent);

      return pp;
   }
   
   /**
    * Get the {@link SearchMode} constant associated with the given mode, used
    * by ADD-SUPER-PROCEDURE calls. If the mode is not one of the SEARCH-SELF 
    * or SEARCH-TARGET values, or an {@link IllegalArgumentException} will be 
    * thrown.
    * <p>
    * The given <code>mode</code> is not compatible with the mode returned by
    * {@link SearchMode#toString()}.
    * 
    * @param    mode
    *           The string representation of the search-mode.
    *           
    * @return   For SEARCH-TARGET value, {@link SearchMode#SEARCH_TARGET} 
    *           will be returned. For SEARCH-SELF value, 
    *           {@link SearchMode#SEARCH_SELF} will be returned.
    *
    * @throws   IllegalArgumentException
    *           If the mode is not of a known value.
    */
   private static SearchMode getSearchMode(String mode)
   {
      if ("SEARCH-SELF".equals(mode))
         return SearchMode.SEARCH_SELF;
      
      if ("SEARCH-TARGET".equals(mode))
         return SearchMode.SEARCH_TARGET;
      
      final String txt = "Unknown mode %s for add-super-procedure calls.";
      throw new IllegalArgumentException(String.format(txt, mode));
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   implements Scopeable
   {
      /**
       * The list of all registered persistent procedures, via RUN PERSISTENT
       * calls.
       */
      private Deque<Object> persistentProcedures = new ArrayDeque<Object>();
      
      /**
       * A map containing the list of super-procedures, for the session,
       * by each instance. 
       */
      private Map<Object, PersistentProc> sessionSuperProcs = 
         new LinkedHashMap<Object, PersistentProc>();

      /** Stack of executed external programs, by instance. */
      private Deque<Object> externalPrograms = new ArrayDeque<Object>();
      
      /** 
       * This map holds the attributes for all registered external procedures,
       * by instance.
       */
      private Map<Object, PersistentProc> ppMap = 
         new HashMap<Object, PersistentProc>();
      
      /** Stack of target-procedures, by instance. */
      private Deque<Object> targetProcedures = new ArrayDeque<Object>();

      /** Stack of source-procedures, by instance. */
      private Deque<Object> sourceProcedures = new ArrayDeque<Object>();
      
      /** Stack of this-procedures, by instance. */
      private Deque<Object> thisProcedures = new ArrayDeque<Object>();
      
      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeStart()
      {
         BlockType bt = TransactionManager.getBlockType();

         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION))
         {
            // push this-procedure to the stack and populate other data
            handle hthis = thisProcedureImpl();
            Object referent = hthis.get();
            if (!ppMap.containsKey(referent))
            {
               ppMap.put(referent, new PersistentProc(hthis, null));
            }
            thisProcedures.push(referent);
            
            if (bt.equals(BlockType.EXTERNAL_PROC))
            {
               if (TransactionManager.getNestingLevel() == 2)
               {
                  // if this is the entry-point, set it as target procedure
                  pushTargetProcedure(hthis);
               }
               externalPrograms.push(referent);
            }
         }
      }

      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeFinished()
      {
         BlockType bt = TransactionManager.getBlockType();
         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION))
         {
            // pop this-procedure from the stack and cleanup
            cleanup(thisProcedures.pop());
         }

         if (!bt.equals(BlockType.EXTERNAL_PROC))
            return;

          Object referent = externalPrograms.pop();
          cleanup(referent);
      }
      
      /**
       * Get the map of session's super-procedures.
       */
      public Map<Object, PersistentProc> getSessionSuperProcedures()
      {
         return getSuperProcedures(null);
      }

      /**
       * Get the map of the super-procedures for the given referent.
       * 
       * @param    referent
       *           The external program instance for which the super-procedures
       *           are needed.
       */
      public Map<Object, PersistentProc> getSuperProcedures(Object referent)
      {
         return referent == null ? sessionSuperProcs 
                                 : ppMap.get(referent).superProcedures;
      }
      
      /**
       * Attempt cleanup of the given external program instance. If is no
       * longer used as a persistent procedure, super procedure, or as a 
       * system handle, the referent will be removed from the {@link #ppMap}
       * map.
       * 
       * @param    referent
       *           The external program instance for which needs to be cleaned
       *           up.
       */
      public void cleanup(Object referent)
      {
         // if no longer used, remove from global map
         if (!systemHandle(referent)                              &&
             !persistentProcedures.contains(referent)             &&
             !sessionSuperProcs.containsKey(referent)             &&
             !superProcForExtProc(referent, persistentProcedures) &&
             !superProcForExtProc(referent, sessionSuperProcs.keySet()))
         {
            ppMap.remove(referent);
         }
      }
      
      /**
       * Check if the given referent is a super-procedure of one of the
       * procedures in the given list.
       * 
       * @param    referent
       *           The external program instance.
       * @param    progs
       *           The list of external programs for which their 
       *           super-procedures list will be searched for the referent.
       *            
       * @return   <code>true</code> if it was found as a super-procedure for
       *           one of the given procedures.
       */
      private boolean superProcForExtProc(Object referent, 
                                          Collection<Object> progs)
      {
         for (Object prog : progs)
         {
            PersistentProc pp = ppMap.get(prog);
            for (Object sreferent : pp.superProcedures.keySet())
            {
               if (sreferent == referent)
               {
                  return true;
               }
            }
         }
         
         return false;
      }
   }

   /**
    * Container for various procedure data.
    */
   private static class PersistentProc
   {
      /**
       * A handle instance for this procedure, so that other handles can
       * be instantiated from it.
       */
      private handle phandle = null;
      
      /**
       * The list of super-procedures for this procedure, by external program
       * instance.
       */
      private Map<Object, PersistentProc> superProcedures = 
         new LinkedHashMap<Object, PersistentProc>();
      
      /** The current search-mode for this external program. */
      private SearchMode searchMode = null;
      
      /** The legacy 4GL name of this external program. */
      private character pname = null;
      
      /** The private data of this procedure handle. */
      private character privateData = null;
      
      /** 
       * The map of name of functions defined IN handle in this external 
       * program, to their {@link Resolvable} instance. 
       */
      private Map<String, Resolvable> inHandleFunctions = 
         new HashMap<String, Resolvable>();
      
      /** A stack of used super-procedures for this procedure handle. */
      private Deque<Object> usedSuperProcedures = new ArrayDeque<Object>();
      
      /**
       * Create a new {@link PersistentProc} instance using the given 
       * procedure handle and search-mode.
       * 
       * @param    phandle
       *           The procedure handle.
       * @param    searchMode
       *           The search-mode. May be null.
       */
      PersistentProc(handle phandle, SearchMode searchMode)
      {
         this.phandle = new handle(phandle);
         this.searchMode = searchMode;
         // at this point, we know the phandle references a valid procedure, 
         // so the name can be retrieved safely.
         this.pname = getFileName(phandle.get());
      }
      
      /**
       * Check if the search-mode for this external program is SEARCH-SELF.
       * 
       * @return   See above.
       */
      boolean isSearchSelf()
      {
         return searchMode == null || searchMode == SearchMode.SEARCH_SELF;
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Initializes the work area, the first time it is requested within a
       * new context. Registers the {@link WorkArea} instance with the
       * {@link TransactionManager}, for scope notifications.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();
         TransactionManager.registerScopeable(wa);
         return wa;
      }
   }
}